@echo off
@set GPU_IDS=%1
@set PYTHON_INTERPRETER=%2
@set SCRIPT_PATH=%3
@set CONFIG_PATH=%4

@set CUDA_VISIBLE_DEVICES=%GPU_IDS%

@REM echo %PYTHON_INTERPRETER% %SCRIPT_PATH% %CONFIG_PATH%
%PYTHON_INTERPRETER% %SCRIPT_PATH% %CONFIG_PATH%